#!/usr/bin/env python

"""
XML-RPC support using libxml2dom.

See: http://www.xmlrpc.com/spec

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The sending and receiving of XML-RPC messages can be done using traditional HTTP
libraries.

See tests/xmlrpc_test.py for more details.
"""

import libxml2dom
from libxml2dom.macrolib import *
from libxml2dom.macrolib import \
    createDocument as Node_createDocument
import datetime

class XMLRPCImplementation(libxml2dom.Implementation):

    "Contains an XML-RPC-specific implementation."

    # Wrapping of documents.

    def adoptDocument(self, node):
        return XMLRPCDocument(node, self)

    # Factory functions.

    def get_node(self, _node, context_node):

        """
        Get a libxml2dom node for the given low-level '_node' and libxml2dom
        'context_node'.
        """

        if Node_nodeType(_node) == context_node.ELEMENT_NODE:

            # Make special elements.

            if Node_localName(_node) in ("methodCall", "methodResponse"):
                return XMLRPCMethodElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "methodName":
                return XMLRPCMethodNameElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "fault":
                return XMLRPCFaultElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "string":
                return XMLRPCStringElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) in ("int", "i4"):
                return XMLRPCIntegerElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "boolean":
                return XMLRPCBooleanElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "double":
                return XMLRPCDoubleElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "dateTime.iso8601":
                return XMLRPCDateTimeElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "base64":
                return XMLRPCBase64Element(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "struct":
                return XMLRPCStructElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "member":
                return XMLRPCMemberElement(_node, self, context_node.ownerDocument)
            elif Node_localName(_node) == "value":
                return XMLRPCValueElement(_node, self, context_node.ownerDocument)

            # Otherwise, make generic XML-RPC elements.

            return XMLRPCElement(_node, self, context_node.ownerDocument)

        else:
            return libxml2dom.Implementation.get_node(self, _node, context_node)

    # Convenience functions.

    def createXMLRPCMessage(self, namespaceURI, localName):

        "Create a new XML-RPC message document (fragment)."

        return XMLRPCDocument(Node_createDocument(namespaceURI, localName, None), self).documentElement

    def createMethodCall(self):
        return self.createXMLRPCMessage(None, "methodCall")

    def createMethodResponse(self):
        return self.createXMLRPCMessage(None, "methodResponse")

# Internal utility functions.

def boolean(s):
    if s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    else:
        raise ValueError, "String value %s not convertable to boolean." % repr(s)

def iso8601(s):
    return s

# Node classes.

class XMLRPCNode(libxml2dom.Node):

    "Convenience modifications to nodes specific to libxml2dom.xmlrpc."

    pass

class XMLRPCElement(XMLRPCNode):

    "An XML-RPC element."

    pass

class XMLRPCDocument(libxml2dom._Document, XMLRPCNode):

    "An XML-RPC document fragment."

    def _method(self):
        return (self.xpath("./methodCall|./methodResponse") or [None])[0]

    method = property(_method)

    # Convenience methods and properties.

    def _methodName(self):
        if self.method is not None:
            return self.method.name
        else:
            return None

    def _parameterItems(self):
        if self.method is not None:
            return self.method.parameterItems
        else:
            return None

    def _parameterValues(self):
        if self.method is not None:
            return self.method.parameterValues
        else:
            return None

    def _parameterMap(self):
        if self.method is not None:
            return self.method.parameterMap
        else:
            return None

    def _fault(self):
        if self.method is not None:
            return self.method.fault
        else:
            return None

    # Node construction methods.

    def createMethodCall(self):
        return self.ownerDocument.createElement("methodCall")

    def createMethodResponse(self):
        return self.ownerDocument.createElement("methodResponse")

    methodName = property(_methodName)
    parameterItems = property(_parameterItems)
    parameterValues = property(_parameterValues)
    parameterMap = property(_parameterMap)
    fault = property(_fault)

class XMLRPCMethodElement(XMLRPCNode):

    "An XML-RPC method element."

    def _fault(self):
        return (self.xpath("./fault") or [None])[0]

    def _methodName(self):
        return (self.xpath("./methodName") or [None])[0]

    def _name(self):
        name = self.methodName
        if name is not None:
            return name.value
        else:
            return None

    def _setName(self, name):
        if self.methodName is None:
            methodName = self.createMethodName()
            self.appendChild(methodName)
        self.methodName.value = name

    def _parameters(self):
        return self.xpath("./params/param")

    def _parameterItems(self):
        values = self.xpath("./params/param/value")
        if values:
            items = []
            for value in values:
                items.append(self._item_contents(value))
            return items
        else:
            return []

    def _parameterValues(self):
        return self._value_contents(self.parameterItems)

    def _setParameterValues(self, parameters):
        param_list = self.parameters
        params = (self.xpath("./params") or [None])[0]
        if params:
            if param_list:
                for param in param_list:
                    params.removeChild(param)
        else:
            params = self.createParameters()
            self.appendChild(params)

        for parameter in parameters:
            param = self.ownerDocument.createElement("param")
            params.appendChild(param)
            value = self.ownerDocument.createElement("value")
            param.appendChild(value)

            # NOTE: Only handles simple types.

            container = self._make_container(parameter)
            value.appendChild(container)
            text = self.ownerDocument.createTextNode(unicode(parameter))
            container.appendChild(text)

    def _parameterMap(self):
        return self._map_contents(self.parameterItems)

    # Internal data.

    converters = {
        "string" : unicode,
        "int" : int,
        "i4" : int,
        "double" : float,
        "boolean" : boolean, # see the module globals
        "dateTime.iso8601" : iso8601, # see the module globals
        "base64" : str
        }

    # Internal methods.

    def _make_container(self, parameter):
        if isinstance(parameter, (str, unicode)):
            return self.ownerDocument.createElement("string")
        elif isinstance(parameter, (int, long)):
            return self.ownerDocument.createElement("int")
        elif isinstance(parameter, float):
            return self.ownerDocument.createElement("double")
        elif isinstance(parameter, bool):
            return self.ownerDocument.createElement("boolean")
        elif isinstance(parameter, datetime.datetime):
            return self.ownerDocument.createElement("dateTime.iso8601")

    def _from_string(self, typename, value):
        return self.converters.get(typename, str)(value)

    def _item_contents(self, value):
        if value.type == "struct":
            items = []
            for member in value.container.members:
                if member.value.type == "struct":
                    items.append((member.name, self._item_contents(member.value)))
                else:
                    items.append((member.name, self._from_string(member.value.type, member.value.container.value)))
            return None, items
        else:
            return None, self._from_string(value.type, value.container.value)

    def _value_contents(self, items):
        values = []
        for name, value in items:
            if isinstance(value, list):
                values.append(self._value_contents(value))
            elif name is None:
                values.append(value)
            else:
                values.append((name, value))
        return values

    def _map_contents(self, items):
        d = {}
        for n, (name, value) in enumerate(items):
            key_name = name or str(n)
            if isinstance(value, list):
                d[key_name] = self._map_contents(value)
            else:
                d[key_name] = value
        return d

    # Node construction methods.

    def createMethodName(self):
        return self.ownerDocument.createElement("methodName")

    def createParameters(self):
        return self.ownerDocument.createElement("params")

    def createFault(self):
        return self.ownerDocument.createElement("fault")

    fault = property(_fault)
    name = property(_name, _setName)
    methodName = property(_methodName)
    parameters = property(_parameters)
    parameterItems = property(_parameterItems)
    parameterValues = property(_parameterValues, _setParameterValues)
    parameterMap = property(_parameterMap)

class XMLRPCStringElement(XMLRPCNode):

    "An XML-RPC string element."

    def _value(self):
        return self.textContent.strip()

    def _setValue(self, value):
        for node in self.childNodes:
            self.removeChild(node)
        text = self.ownerDocument.createTextNode(value)
        self.appendChild(text)

    value = property(_value, _setValue)

class XMLRPCValueElement(XMLRPCStringElement):

    "An XML-RPC value element."

    def _type(self):
        elements = self.xpath("*")
        if elements:
            return elements[0].localName
        else:
            return "string"

    def _container(self):
        return (self.xpath("*") or [self])[0]

    type = property(_type)
    container = property(_container)

class XMLRPCMethodNameElement(XMLRPCStringElement):

    "An XML-RPC method element."

    pass

class XMLRPCIntegerElement(XMLRPCStringElement):

    "An XML-RPC integer element."

    pass

class XMLRPCBooleanElement(XMLRPCStringElement):

    "An XML-RPC boolean element."

    pass

class XMLRPCDoubleElement(XMLRPCStringElement):

    "An XML-RPC double floating point number element."

    pass

class XMLRPCDateTimeElement(XMLRPCStringElement):

    "An XML-RPC date/time element."

    pass

class XMLRPCBase64Element(XMLRPCStringElement):

    "An XML-RPC integer element."

    pass

class XMLRPCStructElement(XMLRPCNode):

    "An XML-RPC structure element."

    def _members(self):
        return self.xpath("./member")

    members = property(_members)

class XMLRPCMemberElement(XMLRPCNode):

    "An XML-RPC structure member element."

    def _name(self):
        value = self.xpath("./name")
        if value:
            return value[0].textContent.strip()
        else:
            return None

    def _value(self):
        return (self.xpath("./value") or [None])[0]

    name = property(_name)
    value = property(_value)

class XMLRPCFaultElement(XMLRPCNode):

    "An XML-RPC fault element."

    def _code(self):
        code = self.xpath("./value/struct/member[./name/text() = 'faultCode']/value/int")
        if code:
            return code[0].value
        else:
            return None

    def _reason(self):
        reason = self.xpath("./value/struct/member[./name/text() = 'faultString']/value/string")
        if reason:
            return reason[0].value
        else:
            return None

    code = property(_code)
    reason = property(_reason)

# Utility functions.

createDocument = libxml2dom.createDocument
createDocumentType = libxml2dom.createDocumentType

def createXMLRPCMessage(namespaceURI, localName):
    return default_impl.createXMLRPCMessage(None, localName)

def createMethodCall():
    return default_impl.createMethodCall()

def createMethodResponse():
    return default_impl.createMethodResponse()

def parse(stream_or_string, html=0, htmlencoding=None, unfinished=0, impl=None):
    return libxml2dom.parse(stream_or_string, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))

def parseFile(filename, html=0, htmlencoding=None, unfinished=0, impl=None):
    return libxml2dom.parseFile(filename, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))

def parseString(s, html=0, htmlencoding=None, unfinished=0, impl=None):
    return libxml2dom.parseString(s, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))

def parseURI(uri, html=0, htmlencoding=None, unfinished=0, impl=None):
    return libxml2dom.parseURI(uri, html=html, htmlencoding=htmlencoding, unfinished=unfinished, impl=(impl or default_impl))

# Single instance of the implementation.

default_impl = XMLRPCImplementation()

# vim: tabstop=4 expandtab shiftwidth=4
